
@file:Suppress("unused")

package ru.dbotthepony.mc.otm.capability

import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.compat.mekanism.Mattery2MekanismEnergyWrapper
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.ifHas
import ru.dbotthepony.mc.otm.core.map
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.tagNotNull

private enum class EnergyFlow {
	INPUT, OUTPUT, BI_DIRECTIONAL
}

sealed class ItemEnergyStorageImpl(
	private val direction: EnergyFlow,
	protected val itemStack: ItemStack,
	maxBatteryLevel: ImpreciseFraction,
	maxInput: ImpreciseFraction?,
	maxOutput: ImpreciseFraction?,
	val initialBatteryLevel: ImpreciseFraction = ImpreciseFraction.ZERO
) : IMatteryEnergyStorage, ICapabilityProvider {
	var maxInput: ImpreciseFraction? = maxInput
		protected set

	var maxOutput: ImpreciseFraction? = maxOutput
		protected set

	private val resolver = LazyOptional.of { this }
	private val resolverMekanism = if (isMekanismLoaded) LazyOptional.of { Mattery2MekanismEnergyWrapper(this) } else null

	override fun <T : Any> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (cap == ForgeCapabilities.ENERGY || cap == MatteryCapability.ENERGY) {
			return resolver.cast()
		} else if (cap == MatteryCapability.MEKANISM_ENERGY) {
			return resolverMekanism?.cast() ?: LazyOptional.empty()
		}

		return LazyOptional.empty()
	}

	override var maxBatteryLevel: ImpreciseFraction = maxBatteryLevel
		protected set

	override var batteryLevel: ImpreciseFraction
		get() = itemStack.tag?.map(NBT_KEY, ImpreciseFraction::deserializeNBT) ?: initialBatteryLevel
		protected set(value) {
			itemStack.tagNotNull[NBT_KEY] = value.serializeNBT()
		}

	override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (direction == EnergyFlow.INPUT)
			return ImpreciseFraction.ZERO

		return extractEnergyInner(howMuch, simulate)
	}

	override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (direction == EnergyFlow.OUTPUT)
			return ImpreciseFraction.ZERO

		return receiveEnergyInner(howMuch, simulate)
	}

	override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive || itemStack.count != 1)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		val maxOutput = maxOutput

		if (maxOutput != null) {
			howMuch = howMuch.coerceAtMost(maxOutput)
		}

		val batteryLevel = batteryLevel

		if (!batteryLevel.isPositive)
			return ImpreciseFraction.ZERO

		val newLevel = (batteryLevel - howMuch).moreThanZero()
		val diff = (batteryLevel - newLevel)

		if (!simulate && batteryLevel != newLevel) {
			this.batteryLevel = newLevel
		}

		return diff
	}

	override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive || itemStack.count != 1)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		val maxInput = maxInput

		if (maxInput != null) {
			howMuch = howMuch.coerceAtMost(maxInput)
		}

		val batteryLevel = batteryLevel

		if (batteryLevel >= maxBatteryLevel)
			return ImpreciseFraction.ZERO

		val newLevel = (batteryLevel + howMuch).coerceAtMost(maxBatteryLevel)
		val diff = (newLevel - batteryLevel)

		if (!simulate && batteryLevel != newLevel) {
			this.batteryLevel = newLevel
		}

		return diff
	}

	override fun canExtract(): Boolean {
		return direction != EnergyFlow.INPUT
	}

	override fun canReceive(): Boolean {
		return direction != EnergyFlow.OUTPUT
	}

	companion object {
		const val NBT_KEY = "energy"
	}
}

open class EnergyConsumerItem(
	stack: ItemStack,
	maxBatteryLevel: ImpreciseFraction,
	maxInput: ImpreciseFraction? = null,
	maxOutput: ImpreciseFraction? = maxInput,
	initialBatteryLevel: ImpreciseFraction = ImpreciseFraction.ZERO
) : ItemEnergyStorageImpl(EnergyFlow.INPUT, stack, maxBatteryLevel, maxInput, maxOutput, initialBatteryLevel)

open class EnergyProducerItem(
	stack: ItemStack,
	maxBatteryLevel: ImpreciseFraction,
	maxInput: ImpreciseFraction? = null,
	maxOutput: ImpreciseFraction? = maxInput,
	initialBatteryLevel: ImpreciseFraction = ImpreciseFraction.ZERO
) : ItemEnergyStorageImpl(EnergyFlow.OUTPUT, stack, maxBatteryLevel, maxInput, maxOutput, initialBatteryLevel)

open class EnergyCapacitorItem(
	stack: ItemStack,
	maxBatteryLevel: ImpreciseFraction,
	maxInput: ImpreciseFraction? = null,
	maxOutput: ImpreciseFraction? = maxInput,
	initialBatteryLevel: ImpreciseFraction = ImpreciseFraction.ZERO
) : ItemEnergyStorageImpl(EnergyFlow.BI_DIRECTIONAL, stack, maxBatteryLevel, maxInput, maxOutput, initialBatteryLevel)

sealed class BlockEnergyStorageImpl constructor(
	protected val listener: () -> Unit,
	private val direction: EnergyFlow,
	maxBatteryLevel: ImpreciseFraction,
	protected var maxInput: ImpreciseFraction?,
	protected var maxOutput: ImpreciseFraction?
) : IMatteryEnergyStorage, INBTSerializable<CompoundTag> {
	override var maxBatteryLevel: ImpreciseFraction = maxBatteryLevel
		protected set

	override var batteryLevel = ImpreciseFraction.ZERO
		protected set

	override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (direction == EnergyFlow.INPUT)
			return ImpreciseFraction.ZERO

		return extractEnergyInner(howMuch, simulate)
	}

	override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (direction == EnergyFlow.OUTPUT)
			return ImpreciseFraction.ZERO

		return receiveEnergyInner(howMuch, simulate)
	}

	override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		val maxOutput = maxOutput

		if (maxOutput != null) {
			howMuch = howMuch.coerceAtMost(maxOutput)
		}

		if (!batteryLevel.isPositive)
			return ImpreciseFraction.ZERO

		val newLevel = (batteryLevel - howMuch).moreThanZero()
		val diff = (batteryLevel - newLevel)

		if (!simulate && batteryLevel != newLevel) {
			batteryLevel = newLevel
			listener.invoke()
		}

		return diff
	}

	override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		val maxInput = maxInput

		if (maxInput != null) {
			howMuch = howMuch.coerceAtMost(maxInput)
		}

		if (batteryLevel >= maxBatteryLevel)
			return ImpreciseFraction.ZERO

		val newLevel = (batteryLevel + howMuch).coerceAtMost(maxBatteryLevel)
		val diff = (newLevel - batteryLevel)

		if (!simulate && batteryLevel != newLevel) {
			batteryLevel = newLevel
			listener.invoke()
		}

		return diff
	}

	override fun canExtract(): Boolean {
		return direction != EnergyFlow.INPUT
	}

	override fun canReceive(): Boolean {
		return direction != EnergyFlow.OUTPUT
	}

	override fun serializeNBT(): CompoundTag {
		val tag = CompoundTag()
		tag["energy_stored"] = batteryLevel.serializeNBT()
		tag["energy_stored_max"] = maxBatteryLevel.serializeNBT()

		maxInput?.let { tag["max_input"] = it.serializeNBT() }
		maxOutput?.let { tag["max_output"] = it.serializeNBT() }

		return tag
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		nbt.ifHas("energy_stored") { batteryLevel = ImpreciseFraction.deserializeNBT(it) }
		nbt.ifHas("energy_stored_max") { maxBatteryLevel = ImpreciseFraction.deserializeNBT(it) }
		nbt.ifHas("max_input") { maxInput = ImpreciseFraction.deserializeNBT(it) }
		nbt.ifHas("max_output") { maxOutput = ImpreciseFraction.deserializeNBT(it) }
	}

	var resolver: LazyOptional<IMatteryEnergyStorage> = LazyOptional.of { this }
		private set

	fun invalidate() {
		resolver.invalidate()
	}

	fun revive() {
		resolver = LazyOptional.of { this }
	}

	companion object {
		val DEFAULT_MAX_IO = ImpreciseFraction(200)
		val DEFAULT_MAX_CAPACITY = ImpreciseFraction(60_000)
	}
}

open class WorkerEnergyStorage(
	listener: () -> Unit,
	maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
	maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction? = maxInput
) : BlockEnergyStorageImpl(listener, EnergyFlow.INPUT, maxBatteryLevel, maxInput, maxOutput) {
	constructor(
		listener: BlockEntity,
		maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
		maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
		maxOutput: ImpreciseFraction? = maxInput) : this({listener.setChanged()}, maxBatteryLevel, maxInput, maxOutput)
}

open class GeneratorEnergyStorage(
	listener: () -> Unit,
	maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
	maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction? = maxInput
) : BlockEnergyStorageImpl(listener, EnergyFlow.OUTPUT, maxBatteryLevel, maxInput, maxOutput) {
	constructor(
		listener: BlockEntity,
		maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
		maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
		maxOutput: ImpreciseFraction? = maxInput) : this({listener.setChanged()}, maxBatteryLevel, maxInput, maxOutput)
}

open class CapacitorEnergyStorage(
	listener: () -> Unit,
	maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
	maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction? = maxInput
) : BlockEnergyStorageImpl(listener, EnergyFlow.BI_DIRECTIONAL, maxBatteryLevel, maxInput, maxOutput) {
	constructor(
		listener: BlockEntity,
		maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
		maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
		maxOutput: ImpreciseFraction? = maxInput) : this({listener.setChanged()}, maxBatteryLevel, maxInput, maxOutput)
}
